// 🥺
#![no_std]
#![feature(default_alloc_error_handler)]
#[macro_use]
extern crate alloc;
use alloc::vec::Vec;

#[cfg(feature = "buddy-alloc")]
mod my_alloc;
mod wasm4;
use wasm4::*;

mod beaste;
mod images;
mod math;
mod maze;
mod random;
mod renderer;

use math::round;
use maze::Board;

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum State {
    Title,
    Maze,
    Level,
    End(bool, bool),
    EndEnd,
}

static mut GAMESTATE: State = State::Title;
static mut LEVEL: u8 = 0;

static mut GLOBAL_MAZE: Board = Board {
    bytes: Vec::new(),
    width: 0,
    height: 0,
    byte_width: 0,
};

static mut FRAME_COUNT: u32 = 0;

static mut PREV: u8 = 0;

static mut PLAYER_X: f32 = 0.0;
static mut PLAYER_Y: f32 = 0.0;
static mut PLAYER_ANGLE: f32 = 0.0;
const TURN_FRAMES: u32 = 20;
const MOVE_FRAMES: u32 = 30;

static mut GOAL: (u8, u8) = (4, 4);

unsafe fn restart_maze() {
    let maze = Board::generate(7 + LEVEL, 10 + LEVEL / 2);
    let player = (maze.random_point().0 as f32, 0.0);
    (PLAYER_X, PLAYER_Y) = player;
    PLAYER_ANGLE = math::PI * 0.5;
    let goal_pos = (maze.random_point().0, (maze.height - 1) / 2 * 2);
    GOAL = goal_pos;
    let beaste_pos = (maze.random_point().0, maze.height / 2);
    beaste::initialize(beaste_pos, LEVEL);
    GLOBAL_MAZE = maze;
}

#[no_mangle]
unsafe fn update() {
    *PALETTE = [0xaea691, 0x887b6a, 0x605444, 0x4e3f2a];
    *DRAW_COLORS = 0x04;
    let old = GAMESTATE;
    random::get_byte();
    GAMESTATE = match GAMESTATE {
        State::Title => update_title(),
        State::Level => update_level(),
        State::Maze => update_maze(),
        State::End(win, retry) => update_end(win, retry),
        State::EndEnd => update_end_end(),
    };
    if GAMESTATE != old && GAMESTATE == State::Maze {
        restart_maze();
    }
    FRAME_COUNT = FRAME_COUNT.wrapping_add(1);
}

unsafe fn update_title() -> State {
    let gamepad = *GAMEPAD1;
    let pressed = gamepad & (gamepad ^ PREV);
    PREV = gamepad;
    text("A Game Against\nThe Beaste", 20, 20);
    text("He can only see\nmovement.", 30, 60);
    text("11 floors deep.", 35, 85);
    text("> Start New Game", 10, 120);

    if pressed & (BUTTON_1 | BUTTON_2) != 0 {
        State::Level
    } else {
        State::Title
    }
}

unsafe fn update_end(did_win: bool, mut retry: bool) -> State {
    if did_win {
        text("  You manage to\ndescend another\nlevel.", 16, 10);
        text("Congratulations.", 24, 70);
        text("  Continue", 30, 120);
    } else {
        text("And the last\none falls.", 50, 10);
        text("Unsurprisingly.", 24, 70);
        text("  Retry", 30, 120);
    }

    text("  Quit", 30, 132);
    let gamepad = *GAMEPAD1;
    let pressed = gamepad & (gamepad ^ PREV);
    PREV = gamepad;
    // retry = retry != (pressed & (BUTTON_UP | BUTTON_DOWN) != 0);
    if pressed & BUTTON_UP != 0 {
        retry = true;
    } else if pressed & BUTTON_DOWN != 0 {
        retry = false;
    }
    let cursor_y = if retry { 120 } else { 132 };
    text(">", 30, cursor_y);

    if gamepad & (BUTTON_1 | BUTTON_2) != 0 {
        if retry {
            if did_win {
                LEVEL += 1;
            }
            State::Level
        } else {
            LEVEL = 0;
            State::Title
        }
    } else {
        State::End(did_win, retry)
    }
}

static mut LEVEL_WAIT: u32 = 0;
unsafe fn update_level() -> State {
    if LEVEL_WAIT == 0 {
        LEVEL_WAIT = FRAME_COUNT + 60;
    }

    text("Level B", 16, 70);
    for i in 0..8 {
        let mask = 128 >> i;
        *DRAW_COLORS = if LEVEL & mask != 0 { 0x43 } else { 0x40 };
        rect(72 + 8 * i as i32, 70, 7, 7);
    }

    if LEVEL_WAIT < FRAME_COUNT {
        LEVEL_WAIT = 0;
        State::Maze
    } else {
        State::Level
    }
}

unsafe fn update_end_end() -> State {
    text("  And so, at the\nbottom of the\nlabyrinth, all the\nlittle people\nscattered as the\ndogs were let\nloose.", 8, 10);
    text(" We pray you\ndon't return.", 24, 120);
    State::EndEnd
}

unsafe fn update_maze() -> State {
    let gamepad = *GAMEPAD1;
    let mut has_moved = false;

    let turn_speed = math::PI / 2.0 / (TURN_FRAMES as f32);
    if gamepad & BUTTON_LEFT != 0 {
        PLAYER_ANGLE -= turn_speed;
        has_moved = true;
    } else if gamepad & BUTTON_RIGHT != 0 {
        PLAYER_ANGLE += turn_speed;
        has_moved = true;
    }

    let move_speed = 1.0 / (MOVE_FRAMES as f32);
    let x_speed = move_speed * math::cos(PLAYER_ANGLE);
    let y_speed = move_speed * math::sin(PLAYER_ANGLE);
    let old_x = PLAYER_X;
    let old_y = PLAYER_Y;
    if gamepad & BUTTON_UP != 0 {
        PLAYER_X += x_speed;
        PLAYER_Y += y_speed;
        has_moved = true;
    } else if gamepad & BUTTON_DOWN != 0 {
        PLAYER_X -= x_speed;
        PLAYER_Y -= y_speed;
        has_moved = true;
    }
    if GLOBAL_MAZE.get_unbounded(round(PLAYER_X) as u8, round(PLAYER_Y) as u8) {
        if !GLOBAL_MAZE.get_unbounded(round(old_x) as u8, round(PLAYER_Y) as u8) {
            PLAYER_X = old_x;
        } else if !GLOBAL_MAZE.get_unbounded(round(PLAYER_X) as u8, round(old_y) as u8) {
            PLAYER_Y = old_y;
        } else {
            PLAYER_X = old_x;
            PLAYER_Y = old_y;
        }
    }

    let pos = (PLAYER_X, PLAYER_Y);
    renderer::draw_maze(&GLOBAL_MAZE, pos, PLAYER_ANGLE);
    let if_moving_pos = if has_moved { Some(pos) } else { None };
    let him_pos = beaste::update(&GLOBAL_MAZE, if_moving_pos);
    *DRAW_COLORS = 4;
    renderer::draw_sprite(
        &images::HIM,
        images::HIM_WIDTH,
        images::HIM_HEIGHT,
        renderer::world_to_local(him_pos, pos, PLAYER_ANGLE),
    );

    let (goal_w, goal_h) = (24, 24);
    let goal_size = goal_w * goal_h / 8;
    let mut goal_sprite = vec![0; goal_size];
    for byte in goal_sprite.iter_mut().take(goal_size) {
        *byte = random::get_byte();
    }
    let goal_float = (GOAL.0 as f32, GOAL.1 as f32);
    *DRAW_COLORS = 1;
    renderer::draw_sprite(
        &goal_sprite,
        goal_w as u32,
        goal_h as u32,
        renderer::world_to_local(goal_float, pos, PLAYER_ANGLE),
    );
    *DRAW_COLORS = 2;

    for w in 0..GLOBAL_MAZE.width {
        for h in 0..GLOBAL_MAZE.height {
            if GLOBAL_MAZE.get(w, h) {
                *DRAW_COLORS = 1;
            } else {
                *DRAW_COLORS = 3;
            }
            pixel(w as i32, h as i32);
        }
    }

    *PALETTE = beaste::get_palette();

    let gx = goal_float.0 - PLAYER_X;
    let gy = goal_float.1 - PLAYER_Y;
    if math::sqrt(gx * gx + gy * gy) < 8.0 {
        if LEVEL == 10 {
            State::EndEnd
        } else {
            State::End(true, true)
        }
    } else if beaste::is_player_eaten(pos) {
        State::End(false, true)
    } else {
        State::Maze
    }
}

#[panic_handler]
#[allow(unused_variables)]
fn panic(info: &core::panic::PanicInfo) -> ! {
    // Can you believe the amount of binary bloat that `format` creates?
    // ...Actually that's not entirely surprising.
    #[cfg(debug_assertions)]
    trace(format!("{}", info));
    loop {}
}
