// American army knife. News. Swen.
#![allow(unused_variables, dead_code, unused_mut)]
const COOL_KNOB: u8 = 0x30;

#[allow(unused_imports)]
use crate::trace;
use crate::{math, random, Vec};

pub type Point = (u8, u8);

#[derive(Debug, Clone)]
pub struct Board {
    pub bytes: Vec<u8>,
    pub width: u8,
    pub height: u8,
    pub byte_width: u8,
}

#[derive(Debug, Copy, Clone)]
pub enum Dir {
    N = 1,
    E = 2,
    W = 4,
    S = 8,
}

impl Dir {
    pub fn iter() -> [Self; 4] {
        [Dir::N, Dir::E, Dir::W, Dir::S]
    }

    pub fn choice(mut n: u8) -> Self {
        debug_assert!(0 < n && n < 16, "n=0b{:08b} not allowed", n);
        let cull = random::get_byte() % n.count_ones() as u8;
        for _ in 0..cull {
            n &= n - 1;
        }
        debug_assert!(0 < n && n < 16, "n={} is bad somehow", n);
        match n & (!n + 1) {
            1 => Dir::N,
            2 => Dir::E,
            4 => Dir::W,
            _ => Dir::S,
        }
    }

    pub fn opposite(&self) -> Self {
        match self {
            Dir::N => Dir::S,
            Dir::E => Dir::W,
            Dir::W => Dir::E,
            Dir::S => Dir::N,
        }
    }

    pub fn travel(&self, from: Point) -> Point {
        let pos_one = 1_u8;
        let neg_one = u8::MAX;
        match self {
            Self::N => (from.0, from.1.wrapping_add(neg_one)),
            Self::E => (from.0.wrapping_add(pos_one), from.1),
            Self::W => (from.0.wrapping_add(neg_one), from.1),
            Self::S => (from.0, from.1.wrapping_add(pos_one)),
        }
    }
}

impl Board {
    pub fn new(width: u8, height: u8) -> Self {
        let byte_width = if width % 8 == 0 {
            width / 8
        } else {
            width / 8 + 1
        };
        Self {
            bytes: vec![0u8; byte_width as usize * height as usize],
            width,
            height,
            byte_width,
        }
    }

    pub fn in_bounds(&self, x: u8, y: u8) -> bool {
        x < self.width && y < self.height
    }

    fn index(&self, x: u8, y: u8) -> (usize, u8) {
        debug_assert!(
            self.in_bounds(x, y),
            "OOB {:?} {:?}",
            (x, y),
            (self.width, self.height)
        );
        let x = x as usize;
        let y = y as usize;
        let width = self.byte_width as usize;

        let byte_index = x / 8 + y * width;
        let mask = 0x01 << (x % 8);
        (byte_index, mask)
    }

    pub fn get(&self, x: u8, y: u8) -> bool {
        let (idx, mask) = self.index(x, y);
        self.bytes[idx] & mask != 0
    }

    pub fn get_unbounded(&self, x: u8, y: u8) -> bool {
        if self.in_bounds(x, y) {
            self.get(x, y)
        } else {
            true
        }
    }

    pub fn set(&mut self, x: u8, y: u8, val: bool) {
        let (idx, mask) = self.index(x, y);
        if val {
            self.bytes[idx] |= mask;
        } else {
            self.bytes[idx] &= !mask;
        }
    }

    pub fn generate(width: u8, height: u8) -> Self {
        let mwidth = (width + 1) / 2;
        let mheight = (height + 1) / 2;
        let mut visited = Self::new(mwidth, mheight);
        let mut ewall = Self::new(mwidth, mheight);
        let mut swall = Self::new(mwidth, mheight);

        let mut points = vec![(0_u8, 0_u8)];
        visited.set(0, 0, true);
        while let Some(&(px, py)) = points.last() {
            let mut neighbors = 0_u8;
            for dir in Dir::iter() {
                let (cx, cy) = dir.travel((px, py));
                if !visited.get_unbounded(cx, cy) {
                    neighbors += dir as u8;
                }
            }
            if neighbors == 0 {
                points.pop();
                continue;
            }
            let next = Dir::choice(neighbors);
            match next {
                Dir::N => swall.set(px, py - 1, true),
                Dir::E => ewall.set(px, py, true),
                Dir::W => ewall.set(px - 1, py, true),
                Dir::S => swall.set(px, py, true),
            }
            let (nx, ny) = next.travel((px, py));
            visited.set(nx, ny, true);
            points.push((nx, ny));
        }
        drop(visited);
        let mut board = Self::new(width, height);
        for x in 0..width {
            for y in 0..height {
                let (xp, yp) = (x % 2, y % 2);
                let cell = if xp == 0 && yp == 0 {
                    false
                } else if xp == 1 && yp == 1 {
                    true
                } else if yp == 1 {
                    !swall.get(x / 2, y / 2)
                } else if random::get_byte() < COOL_KNOB {
                    false
                } else {
                    !ewall.get(x / 2, y / 2)
                };
                board.set(x, y, cell);
            }
        }
        for x in 0..mwidth {
            for y in 0..mheight {
                let (px, py) = (x * 2 + 1, y * 2 + 1);
                let mut is_there = true;
                for dir in Dir::iter() {
                    let (cx, cy) = dir.travel((px, py));
                    if board.get_unbounded(cx, cy) {
                        is_there = false;
                        break;
                    }
                }
                if is_there {
                    match (random::get_bit(), random::get_bit()) {
                        (true, true) => board.set(px + 1, py, true),
                        (true, false) => board.set(px, py + 1, true),
                        (false, true) => board.set(px - 1, py, true),
                        (false, false) => board.set(px, py - 1, true),
                    }
                }
            }
        }
        board
    }

    pub fn nav(&self, from: Point, to: Point) -> Dir {
        let mut bounds = self.clone();
        let mut buddies: Vec<Point> = vec![to];
        loop {
            let mut offspring = Vec::<Point>::new();
            for &b in &buddies {
                bounds.set(b.0, b.1, true);
                for d in Dir::iter() {
                    let next = d.travel(b);
                    if next == from {
                        return d.opposite();
                    } else if !bounds.get_unbounded(next.0, next.1) {
                        offspring.push(next);
                    }
                }
            }
            if offspring.is_empty() {
                break;
            }
            // Distinct possibility of duplicate offspring.
            buddies = offspring;
        }
        // Bad.
        debug_assert!(false, "can't reach goal: {:?}", buddies);
        for dir in Dir::iter() {
            let (cx, cy) = dir.travel(from);
            if !self.get_unbounded(cx, cy) {
                return dir;
            }
        }
        // Really bad.
        Dir::N
    }

    pub fn can_see(&self, from: (f32, f32), to: (f32, f32)) -> bool {
        let x = to.0 - from.0;
        let y = to.1 - from.1;
        let target_dist = math::sqrt(x * x + y * y);
        if target_dist < 0.3 {
            return true;
        }
        let eye: (f32, f32) = (x / target_dist, y / target_dist);
        let mut ray_dist = 0.0;
        while ray_dist < target_dist {
            ray_dist += 0.01;
            let test_x = math::round(from.0 + eye.0 * ray_dist) as u8;
            let test_y = math::round(from.1 + eye.1 * ray_dist) as u8;
            if self.get_unbounded(test_x, test_y) {
                return false;
            }
        }
        true
    }

    pub fn random_point(&self) -> Point {
        let x = random::get_word() % (self.width as u16);
        let y = random::get_word() % (self.height as u16);
        (x as u8 / 2 * 2, y as u8 / 2 * 2)
    }
}
